# https://github.com/python-trio/outcome/tree/6a3192f306ead4900a33fa8c47e5af5430e37692
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
TODO:
 - hide Value and Error classes like Maybes
"""
import abc
import dataclasses as dc
import typing as ta

from ..lite.abstract import Abstract


T = ta.TypeVar('T')
ValueT_co = ta.TypeVar('ValueT_co', covariant=True)
ResultT = ta.TypeVar('ResultT')
ArgsT = ta.ParamSpec('ArgsT')


##


class OutcomeAlreadyUnwrappedError(RuntimeError):
    """An Outcome can only be unwrapped once."""


def _remove_tb_frames(exc: BaseException, n: int) -> BaseException:
    tb: ta.Any = exc.__traceback__
    for _ in range(n):
        if tb is None:
            raise RuntimeError
        tb = tb.tb_next
    return exc.with_traceback(tb)


##


@ta.overload
def capture(
        sync_fn: ta.Callable[ArgsT, ta.NoReturn],
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
) -> 'Error':
    ...


@ta.overload
def capture(
        sync_fn: ta.Callable[ArgsT, ResultT],
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
) -> ta.Union['Value[ResultT]', 'Error']:
    ...


def capture(
        sync_fn: ta.Callable[ArgsT, ResultT],
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
) -> ta.Union['Value[ResultT]', 'Error']:
    """
    Run ``sync_fn(*args, **kwargs)`` and capture the result.

    Returns:
      Either a :class:`Value` or :class:`Error` as appropriate.
    """

    try:
        return Value(sync_fn(*args, **kwargs))

    except BaseException as exc:  # noqa
        exc = _remove_tb_frames(exc, 1)
        return Error(exc)


#


@ta.overload
async def acapture(
        async_fn: ta.Callable[ArgsT, ta.Awaitable[ta.NoReturn]],
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
) -> 'Error':
    ...


@ta.overload
async def acapture(
        async_fn: ta.Callable[ArgsT, ta.Awaitable[ResultT]],
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
) -> ta.Union['Value[ResultT]', 'Error']:
    ...


async def acapture(
        async_fn: ta.Callable[ArgsT, ta.Awaitable[ResultT]],
        *args: ArgsT.args,
        **kwargs: ArgsT.kwargs,
) -> ta.Union['Value[ResultT]', 'Error']:
    """
    Run ``await async_fn(*args, **kwargs)`` and capture the result.

    Returns:
      Either a :class:`Value` or :class:`Error` as appropriate.
    """

    try:
        return Value(await async_fn(*args, **kwargs))

    except BaseException as exc:  # noqa
        exc = _remove_tb_frames(exc, 1)
        return Error(exc)


##


@dc.dataclass(repr=False, init=False, slots=True, frozen=True, order=True)
class Outcome(Abstract, ta.Generic[ValueT_co]):
    """
    An abstract class representing the result of a Python computation.

    This class has two concrete subclasses: :class:`Value` representing a value, and :class:`Error` representing an
    exception.

    In addition to the methods described below, comparison operators on :class:`Value` and :class:`Error` objects
    (``==``, ``<``, etc.) check that the other object is also a :class:`Value` or :class:`Error` object respectively,
    and then compare the contained objects.

    :class:`Outcome` objects are hashable if the contained objects are hashable.
    """

    @property
    @abc.abstractmethod
    def is_error(self) -> bool:
        raise NotImplementedError

    def value(self) -> ValueT_co:
        raise TypeError

    def error(self) -> BaseException:
        raise TypeError

    #

    _unwrapped: bool = dc.field(default=False, compare=False, init=False)

    def _set_unwrapped(self) -> None:
        if self._unwrapped:
            raise OutcomeAlreadyUnwrappedError
        object.__setattr__(self, '_unwrapped', True)

    @abc.abstractmethod
    def unwrap(self) -> ValueT_co:
        """
        Return or raise the contained value or exception.

        These two lines of code are equivalent::

           x = fn(*args)
           x = outcome.capture(fn, *args).unwrap()
        """

    @abc.abstractmethod
    def send(self, gen: ta.Generator[ResultT, ValueT_co, object]) -> ResultT:
        """
        Send or throw the contained value or exception into the given generator object.

        Args:
          gen: A generator object supporting ``.send()`` and ``.throw()`` methods.
        """

    @abc.abstractmethod
    async def asend(self, agen: ta.AsyncGenerator[ResultT, ValueT_co]) -> ResultT:
        """
        Send or throw the contained value or exception into the given async generator object.

        Args:
          agen: An async generator object supporting ``.asend()`` and ``.athrow()`` methods.
        """


#


@ta.final
@dc.dataclass(frozen=True, repr=False, slots=True, order=True)
class Value(Outcome[ValueT_co], ta.Generic[ValueT_co]):
    """Concrete :class:`Outcome` subclass representing a regular value."""

    _value: ValueT_co  # type: ignore[misc]  # FIXME: Cannot use a covariant type variable as a parameter

    def __repr__(self) -> str:
        return f'Value({self._value!r})'

    #

    @property
    def is_error(self) -> bool:
        return True

    def value(self) -> ValueT_co:
        return self._value

    #

    def unwrap(self) -> ValueT_co:
        self._set_unwrapped()
        return self._value

    def send(self, gen: ta.Generator[ResultT, ValueT_co, object]) -> ResultT:
        self._set_unwrapped()
        return gen.send(self._value)

    async def asend(self, agen: ta.AsyncGenerator[ResultT, ValueT_co]) -> ResultT:
        self._set_unwrapped()
        return await agen.asend(self._value)


def value(v: T) -> Outcome[T]:
    return Value(v)


#


@ta.final
@dc.dataclass(frozen=True, repr=False, slots=True, order=True)
class Error(Outcome[ta.NoReturn]):
    """Concrete :class:`Outcome` subclass representing a raised exception."""

    _error: BaseException

    def __post_init__(self) -> None:
        if not isinstance(self._error, BaseException):
            raise TypeError(self._error)

    def __repr__(self) -> str:
        return f'Error({self._error!r})'

    #

    @property
    def is_error(self) -> bool:
        return True

    def error(self) -> BaseException:
        return self._error

    #

    def unwrap(self) -> ta.NoReturn:
        self._set_unwrapped()

        # Tracebacks show the 'raise' line below out of context, so let's give this variable a name that makes sense out
        # of context.
        captured_error = self._error

        try:
            raise captured_error

        finally:
            # We want to avoid creating a reference cycle here. Python does collect cycles just fine, so it wouldn't be
            # the end of the world if we did create a cycle, but the cyclic garbage collector adds latency to Python
            # programs, and the more cycles you create, the more often it runs, so it's nicer to avoid creating them in
            # the first place. For more details see:
            #
            #    https://github.com/python-trio/trio/issues/1770
            #
            # In particular, by deleting this local variables from the 'unwrap' methods frame, we avoid the
            # 'captured_error' object's __traceback__ from indirectly referencing 'captured_error'.
            del captured_error, self

    def send(self, gen: ta.Generator[ResultT, ta.NoReturn, object]) -> ResultT:
        self._set_unwrapped()
        return gen.throw(self._error)

    async def asend(self, agen: ta.AsyncGenerator[ResultT, ta.NoReturn]) -> ResultT:
        self._set_unwrapped()
        return await agen.athrow(self._error)


def error(e: BaseException) -> Outcome[T]:
    return Error(e)


##


# A convenience alias to a union of both results, allowing exhaustiveness checking.
Either = Value[ValueT_co] | Error



# MIT License
#
# Copyright (c) 2023 Joshua George Albert
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# https://github.com/Joshuaalbert/FairAsyncRLock/blob/7c3ed4e08892638dba262eaf0b1a7b1fad4b2608/fair_async_rlock/fair_async_rlock.py
import asyncio
import collections
import typing as ta


##


class AsyncioRLock:
    def __init__(self) -> None:
        super().__init__()

        self._owner: asyncio.Task | None = None
        self._count = 0
        self._owner_transfer = False
        self._queue: collections.deque[asyncio.Event] = collections.deque()

    def is_owner(self, task: asyncio.Task | None = None) -> bool:
        if task is None:
            task = asyncio.current_task()
        return self._owner == task

    def is_locked(self) -> bool:
        return self._owner is not None

    async def acquire(self) -> None:
        me = asyncio.current_task()

        # If the lock is reentrant, acquire it immediately
        if self.is_owner(task=me):
            self._count += 1
            return

        # If the lock is free (and ownership not in midst of transfer), acquire it immediately
        if not self._count and not self._owner_transfer:
            self._owner = me
            self._count = 1
            return

        # Create an event for this task, to notify when it's ready for acquire
        event = asyncio.Event()
        self._queue.append(event)

        # Wait for the lock to be free, then acquire
        try:
            await event.wait()
            self._owner_transfer = False
            self._owner = me
            self._count = 1

        except asyncio.CancelledError:
            try:  # if in queue, then cancelled before release
                self._queue.remove(event)

            except ValueError:  # otherwise, release happened, this was next, and we simulate passing on
                self._owner_transfer = False
                self._owner = me
                self._count = 1
                self._current_task_release()

            raise

    def _current_task_release(self) -> None:
        self._count -= 1
        if not self._count:
            self._owner = None

            if self._queue:
                # Wake up the next task in the queue
                event = self._queue.popleft()
                event.set()

                # Setting this here prevents another task getting lock until owner transfer.
                self._owner_transfer = True

    def release(self) -> None:
        me = asyncio.current_task()

        if self._owner is None:
            raise RuntimeError(f'Cannot release un-acquired lock. {me} tried to release.')

        if not self.is_owner(task=me):
            raise RuntimeError(f'Cannot release foreign lock. {me} tried to unlock {self._owner}.')

        self._current_task_release()

    async def __aenter__(self) -> ta.Self:
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.release()



def get_bit(bit: int, value: int) -> int:
    return (value >> bit) & 1


def get_bits(bits_from: int, num_bits: int, value: int) -> int:
    return (value & ((1 << (bits_from + num_bits)) - 1)) >> bits_from


def set_bit(bit: int, bit_value: int, value: int) -> int:
    if bit_value:
        return value | (1 << bit)
    else:
        return value & ~(1 << bit)


def set_bits(bits_from: int, num_bits: int, bits_value: int, value: int) -> int:
    return value & ~(((1 << num_bits) - 1) << bits_from) | (bits_value << bits_from)



import typing as ta

import sqlalchemy as sa

from ... import check
from .. import api
from ..api.dbapi import build_dbapi_columns


T = ta.TypeVar('T')


##


class SqlalchemyApiWrapper(api.ContextCloser, ta.Generic[T]):
    def __init__(
            self,
            u: T,
            *,
            auto_close: bool = False,
    ) -> None:
        super().__init__()

        self._u = u
        self._auto_close = auto_close

    @property
    @ta.override
    def _is_resourceless(self) -> bool:
        return not self._auto_close

    def _close(self) -> None:
        if self._auto_close and hasattr(self._u, 'close'):
            self._u.close()

        super()._close()


##


class SqlalchemyApiRows(api.Rows):
    def __init__(self, columns: api.Columns, rows: ta.Sequence[api.Row]) -> None:
        super().__init__()

        self._columns = columns
        self._rows = rows

        self._it = iter(self._rows)

    @property
    def columns(self) -> api.Columns:
        return self._columns

    def __next__(self) -> api.Row:
        return next(self._it)


class SqlalchemyApiConn(SqlalchemyApiWrapper[sa.engine.Connection], api.Conn):
    @property
    def adapter(self) -> api.Adapter:
        raise NotImplementedError

    def query(self, query: api.Query) -> api.Rows:
        check.empty(query.args)
        result: sa.engine.cursor.CursorResult
        with self._u.execute(sa.text(query.text)) as result:
            cols = build_dbapi_columns(result.cursor.description)
            sa_rows = result.fetchall()
            rows = [
                api.Row(cols, tuple(sa_row))
                for sa_row in sa_rows
            ]
        return SqlalchemyApiRows(cols, rows)


class SqlalchemyApiDb(SqlalchemyApiWrapper[sa.engine.Engine], api.Db):
    def connect(self) -> api.Conn:
        return SqlalchemyApiConn(self._u.connect(), auto_close=True)

    @property
    def adapter(self) -> api.Adapter:
        raise NotImplementedError


class SqlalchemyApiAdapter(api.Adapter):
    def scan_type(self, c: api.Column) -> type:
        raise NotImplementedError


##


@ta.overload
def api_adapt(o: sa.engine.Connection) -> SqlalchemyApiConn:
    ...


@ta.overload
def api_adapt(o: api.Conn) -> api.Conn:
    ...


@ta.overload
def api_adapt(o: sa.engine.Engine) -> SqlalchemyApiDb:
    ...


@ta.overload
def api_adapt(o: api.Db) -> api.Db:
    ...


def api_adapt(o):
    if isinstance(o, sa.engine.Connection):
        return SqlalchemyApiConn(o)
    elif isinstance(o, api.Conn):
        return o

    elif isinstance(o, sa.engine.Engine):
        return SqlalchemyApiDb(o)
    elif isinstance(o, api.Db):
        return o

    else:
        raise TypeError(o)



"""
TODO:
 - .super(instance[_cls], owner)
 - ALT: A.f(super(), ... ? :/
 - classmethod/staticmethod
"""
import functools
import typing as ta

from .. import check
from .. import collections as col
from .dispatch import Dispatcher
from .impls import get_impl_func_cls_set


T = ta.TypeVar('T')
R = ta.TypeVar('R')
P = ta.ParamSpec('P')


##


class Method(ta.Generic[P, R]):
    """
    MRO-honoring instancemethod singledispatch. There are many ways to do this, and this one is attr name based: a class
    is considered to have method implementations that have been registered to a given method based on whether or not
    they are accessible by a non-shadowed, MRO-resolved named attribute on that class.

    Care must be taken when overriding registered implementations from superclasses in subclasses - shadowing the name
    of the superclass method will not automatically register the new method with the same name to the dispatch method -
    it must be explicitly `@register`'ed itself. This is a feature, allowing for selective de-registration of
    implementations in subclasses via name shadowing.

    Methods can choose to allow external installation of implementations outside of direct subclasses. This is to be
    used *extremely* rarely - basically only in the rare case of externally extensible type hierarchies with visitors.
    """

    def __init__(
            self,
            func: ta.Callable,
            *,
            installable: bool = False,
            requires_override: bool = False,
            instance_cache: bool = False,
    ) -> None:
        super().__init__()

        if not callable(func) and not hasattr(func, '__get__'):  # type: ignore
            raise TypeError(f'{func!r} is not callable or a descriptor')

        self._func = func
        self._installable = installable
        self._instance_cache = instance_cache

        self._registry: col.AttrRegistry[ta.Callable, Method._Entry] = col.AttrRegistry(
            requires_override=requires_override,
        )

        self._cache: col.AttrRegistryCache[ta.Callable, Method._Entry, ta.Callable] = col.AttrRegistryCache(
            self._registry,
            self._prepare,
        )

        # bpo-45678: special-casing for classmethod/staticmethod in Python <=3.9, as functools.update_wrapper doesn't
        # work properly in singledispatchmethod.__get__ if it is applied to an unbound classmethod/staticmethod
        unwrapped_func: ta.Any
        if isinstance(func, (staticmethod, classmethod)):
            unwrapped_func = func.__func__
        else:
            unwrapped_func = func
        self._unwrapped_func = unwrapped_func
        self._is_abstractmethod = getattr(func, '__isabstractmethod__', False)  # noqa
        self.update_wrapper(self)

        self._owner: type | None = None
        self._name: str | None = None

    class _Entry:
        cls_set: frozenset[type]

    def __set_name__(self, owner, name):
        if self._owner is None:
            self._owner = owner
        if self._name is None:
            self._name = name

    def __repr__(self) -> str:
        return f'<{type(self).__module__}.{type(self).__qualname__}:{self._func} at 0x{id(self):x}>'

    def update_wrapper(self, wrapper: T) -> T:
        for attr in functools.WRAPPER_ASSIGNMENTS:
            try:
                value = getattr(self._unwrapped_func, attr)
            except AttributeError:
                continue
            setattr(wrapper, attr, value)
        setattr(wrapper, '__isabstractmethod__', self._is_abstractmethod)  # noqa
        return wrapper

    def register(self, impl: T, cls_set: frozenset[type] | None = None) -> T:
        check.callable(impl)

        entry = Method._Entry()
        if cls_set is not None:
            entry.cls_set = cls_set

        self._registry.register(ta.cast(ta.Callable, impl), entry)

        return impl

    def _build_dispatcher(self, collected: ta.Mapping[str, tuple[ta.Callable, _Entry]]) -> Dispatcher[str]:
        disp: Dispatcher[str] = Dispatcher()

        for a, (f, e) in collected.items():
            try:
                cls_set = e.cls_set
            except AttributeError:
                cls_set = e.cls_set = get_impl_func_cls_set(f, arg_offset=1)

            disp.register(a, cls_set)

        return disp

    def _build_dispatch_func(self, disp: Dispatcher[str]) -> ta.Callable:
        dispatch = disp.dispatch
        type_ = type
        getattr_ = getattr
        base_func = self._func
        func_name = getattr(base_func, '__name__', 'singledispatch method')

        def __call__(self, *args, **kwargs):  # noqa
            if not args:
                raise TypeError(f'{func_name} requires at least 1 positional argument')

            if (impl_att := dispatch(type_(args[0]))) is not None:
                return getattr_(self, impl_att)(*args, **kwargs)

            return base_func.__get__(self)(*args, **kwargs)  # noqa

        self.update_wrapper(__call__)
        return __call__

    def _prepare(self, instance_cls: type, collected: ta.Mapping[str, tuple[ta.Callable, _Entry]]) -> ta.Callable:
        disp = self._build_dispatcher(collected)
        func = self._build_dispatch_func(disp)
        return func

    def __get__(self, instance, owner=None):
        if instance is None:
            # FIXME: classmethod/staticmethod
            return self

        if self._instance_cache:
            try:
                return instance.__dict__[self._name]
            except KeyError:
                pass

        bound = self._cache.get(type(instance)).__get__(instance, owner)  # noqa

        if self._instance_cache:
            instance.__dict__[self._name] = bound

        return bound

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        instance, *rest = args

        # if instance_cls is super:
        #     owner = instance.__self_class__.__mro__[instance.__self_class__.__mro__.index(instance.__thisclass__) + 1]
        #     att_disp = self.build_attr_dispatcher(instance.__self_class__, owner)
        #     func = self.build_dispatch_func(att_disp)
        #     return func.__get__(instance, instance.__thisclass__)(*rest, **kwargs)

        return self.__get__(instance)(*rest, **kwargs)


##


def method(
        *,
        installable: bool = False,
        requires_override: bool = False,
        instance_cache: bool = False,
) -> ta.Callable[[ta.Callable[P, R]], Method[P, R]]:  # noqa
    return functools.partial(
        Method,  # type: ignore[arg-type]
        installable=installable,
        requires_override=requires_override,
        instance_cache=instance_cache,
    )


#


def install_method(
        mth: ta.Any,
        *,
        name: str | None = None,
        on: type | None = None,
) -> ta.Callable[[T], T]:
    mth = check.isinstance(mth, Method)
    if not mth._installable:  # noqa
        raise TypeError(f'Method not installable: {mth}')

    def inner(fn):
        a = name
        if a is None:
            a = fn.__name__

        owner: type = check.not_none(mth._owner)  # noqa
        if on is None:
            cls = owner
        else:
            cls = check.issubclass(on, owner)  # noqa

        check.arg(not hasattr(cls, a))
        setattr(cls, a, fn)

        mth.register(fn)

        return fn

    return inner



"""
TODO:
 - option to coordinate with objects and omit if empty / render unboxed
"""
import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import MarshalFactoryContext
from ..base.contexts import UnmarshalContext
from ..base.contexts import UnmarshalFactoryContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.method import MarshalerFactoryMethodClass
from ..factories.method import UnmarshalerFactoryMethodClass


##


@dc.dataclass(frozen=True)
class MaybeMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        m: lang.Maybe = o
        if m.present:
            return [self.e.marshal(ctx, m.must())]
        else:
            return []


class MaybeMarshalerFactory(MarshalerFactoryMethodClass):
    @MarshalerFactoryMethodClass.make_marshaler.register
    def _make_generic(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, rfl.Generic) and rty.cls is lang.Maybe):
            return None
        return lambda: MaybeMarshaler(ctx.make_marshaler(check.single(rty.args)))

    @MarshalerFactoryMethodClass.make_marshaler.register
    def _make_concrete(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, lang.Maybe)):
            return None
        return lambda: MaybeMarshaler(ctx.make_marshaler(ta.Any))


#


@dc.dataclass(frozen=True)
class MaybeUnmarshaler(Unmarshaler):
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if v:
            return lang.just(self.e.unmarshal(ctx, check.single(v)))  # type: ignore
        else:
            return lang.empty()


class MaybeUnmarshalerFactory(UnmarshalerFactoryMethodClass):
    @UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_generic(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, rfl.Generic) and rty.cls is lang.Maybe):
            return None
        return lambda: MaybeUnmarshaler(ctx.make_unmarshaler(check.single(rty.args)))

    @UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_concrete(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, lang.Maybe)):
            return None
        return lambda: MaybeUnmarshaler(ctx.make_unmarshaler(ta.Any))



"""
FIXME:
 - use wrapping.py with all its todos
"""
import dataclasses as dc
import textwrap
import typing as ta

from ... import lang
from ..textwrap import TextwrapOpts
from .parts import Blank
from .parts import Block
from .parts import Indent
from .parts import List
from .parts import Part
from .parts import Text
from .parts import blockify
from .parts import unblockify
from .utils import all_same


##


class TextReflower(ta.Protocol):
    def __call__(self, strs: ta.Sequence[str], current_indent: int) -> ta.Sequence[str]: ...


class NopReflower:
    def __call__(self, strs: ta.Sequence[str], current_indent: int = 0) -> ta.Sequence[str]:
        return strs


@dc.dataclass(frozen=True)
class JoiningReflower:
    sep: str = ' '

    def __call__(self, strs: ta.Sequence[str], current_indent: int = 0) -> ta.Sequence[str]:
        return [self.sep.join(strs)]


@dc.dataclass(frozen=True)
class TextwrapReflower:
    sep: str = ' '
    opts: TextwrapOpts = TextwrapOpts()

    def __call__(self, strs: ta.Sequence[str], current_indent: int = 0) -> ta.Sequence[str]:
        return textwrap.wrap(
            self.sep.join(strs),
            **dc.asdict(dc.replace(
                self.opts,
                width=self.opts.width - current_indent,
            )),
        )


##


def reflow_block_text(
        root: Part,
        reflower: TextReflower = JoiningReflower(),
) -> Part:
    def rec(p: Part, ci: int) -> Part:
        if isinstance(p, Blank):
            return p

        elif isinstance(p, Text):
            if (rf := reflower([p.s], ci)) != [p.s]:
                p = blockify(*map(Text, rf))  # noqa
            return p

        elif isinstance(p, Indent):
            if (np := rec(p.p, ci + p.n)) is not p.p:
                p = Indent(p.n, np)  # type: ignore[arg-type]
            return p

        elif isinstance(p, List):
            ne = [rec(e, ci + len(p.d) + 1) for e in p.es]
            if not all_same(ne, p.es):
                p = List(p.d, ne)
            return p

        elif not isinstance(p, Block):
            raise TypeError(p)

        ps = [rec(c, ci) for c in p.ps]

        ps = list(unblockify(blockify(*ps)))

        new: list[Part | list[str]] = []
        for c in ps:
            if isinstance(c, Text):
                if new and isinstance(x := new[-1], list):
                    x.append(c.s)
                else:
                    new.append([c.s])
            else:
                new.append(c)

        return blockify(*lang.flatmap(
            lambda x: map(Text, reflower(x, ci)) if isinstance(x, list) else [x],  # noqa
            new,
        ))

    return rec(root, 0)



"""
TODO:
 - StrView, BytesView - in lieu of hkt lol
 - cext? even necessary?
 - __eq__, cmp, __hash__
 - __buffer__
 - optimize `slice(None)`, keep as SeqView but fast path ops
 - shorter repr if __len__ > some threshold
  - use materialize()?
"""
import operator
import typing as ta


T = ta.TypeVar('T')


##


def iterslice(
        seq: ta.Sequence[T],
        slc: slice,
) -> ta.Iterator[T]:
    return map(seq.__getitem__, range(*slc.indices(len(seq))))


def iterrange(
        seq: ta.Sequence[T],
        start: int | None = None,
        stop: int | None = None,
        step: int | None = None,
) -> ta.Iterator[T]:
    return iterslice(seq, slice(start, stop, step))


##


def seqs_all(
        fn: ta.Callable[[T, T], bool],
        l_seq: ta.Sequence[T],
        *r_seqs: ta.Sequence[T],
) -> bool:
    if not all(len(r_seq) == len(l_seq) for r_seq in r_seqs):
        return False

    return all(
        fn(l, r)
        for r_seq in r_seqs
        for l, r in zip(l_seq, r_seq, strict=True)
    )


def seqs_equal(
        l_seq: ta.Sequence[T],
        *r_seqs: ta.Sequence[T],
) -> bool:
    return seqs_all(operator.eq, l_seq, *r_seqs)


def seqs_identical(
        l_seq: ta.Sequence[T],
        *r_seqs: ta.Sequence[T],
) -> bool:
    return seqs_all(operator.is_, l_seq, *r_seqs)


##


@ta.final
class SeqView(ta.Sequence[T]):
    def __init__(self, data: ta.Sequence[T], slice_: slice = slice(None)) -> None:
        if data.__class__ is SeqView:
            self._data = data._data  # type: ignore[attr-defined]  # noqa
            self._range = data._range[slice_]  # type: ignore[attr-defined]  # noqa
        else:
            self._data = data
            self._range = range(*slice_.indices(len(data)))

    def __init_subclass__(cls, **kwargs):
        raise TypeError

    _data: ta.Sequence[T]
    _range: range

    @classmethod
    def _from_range(cls, base: ta.Sequence[T], rng: range) -> 'SeqView[T]':
        self = object.__new__(cls)
        self._data = base
        self._range = rng
        return self

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._data!r}, {self.slice!r})'

    #

    def __len__(self) -> int:
        return len(self._range)

    def __getitem__(self, key: int | slice) -> ta.Union[T, 'SeqView[T]']:  # type: ignore[override]
        if isinstance(key, slice):
            nr = self._range[key]
            return SeqView._from_range(self._data, nr)
        return self._data[self._range[key]]

    def __iter__(self) -> ta.Iterator[T]:
        return map(self._data.__getitem__, self._range)

    def __reversed__(self) -> ta.Iterator[T]:
        return map(self._data.__getitem__, reversed(self._range))

    def count(self, value: ta.Any) -> int:
        c = 0
        for i in self._range:
            if self._data[i] == value:
                c += 1
        return c

    def index(self, value: ta.Any, start: int = 0, stop: int | None = None) -> int:
        sub = self._range[slice(start, stop, 1)]
        for off, i in enumerate(sub):
            if self._data[i] == value:
                return off
        raise ValueError(f'{value!r} is not in view')

    #

    @property
    def data(self) -> ta.Sequence[T]:
        return self._data

    _slice: slice

    @property
    def slice(self) -> slice:
        try:
            return self._slice
        except AttributeError:
            pass

        step = self._range.step
        start = self._range.start
        if len(self._range) == 0:
            stop = start
        else:
            last = start + (len(self._range) - 1) * step
            stop = last + (1 if step > 0 else -1)
        slc = slice(start, stop, step)

        self._slice = slc
        return slc

    def materialize(self) -> ta.Sequence[T]:
        return self._data[self.slice]



"""
TODO:
 - md5 spec to fn name
"""
import abc
import dataclasses as dc
import typing as ta

from .... import check
from .... import lang
from ..utils import repr_round_trip_value
from .globals import FN_GLOBAL_IMPORTS
from .globals import FN_GLOBALS
from .globals import FUNCTION_TYPE_GLOBAL
from .globals import PROPERTY_GLOBAL
from .globals import TYPE_ERROR_GLOBAL
from .idents import CLS_IDENT
from .idents import IDENT_PREFIX
from .ops import AddMethodOp
from .ops import AddPropertyOp
from .ops import IfAttrPresent
from .ops import Op
from .ops import OpRef
from .ops import Ref
from .ops import SetAttrOp
from .ops import get_op_refs


T = ta.TypeVar('T')


##


class OpCompiler:
    class Style(lang.Abstract):
        @abc.abstractmethod
        def header_lines(self) -> ta.Sequence[str]:
            raise NotImplementedError

        @abc.abstractmethod
        def globals_ns(self) -> ta.Mapping[str, ta.Any]:
            raise NotImplementedError

    class JitStyle(Style):
        def header_lines(self) -> ta.Sequence[str]:
            return []

        def globals_ns(self) -> ta.Mapping[str, ta.Any]:
            return dict(FN_GLOBAL_IMPORTS)

    class AotStyle(Style):
        HEADER_LINES: ta.ClassVar[ta.Sequence[str]] = [
            '# type: ignore',
            '# ruff: noqa',
            '# flake8: noqa',
            '# @omlish-generated',
        ]

        def header_lines(self) -> ta.Sequence[str]:
            return [
                *self.HEADER_LINES,
                *[
                    f'import {i}'
                    for i in FN_GLOBAL_IMPORTS
                ],
            ]

        def globals_ns(self) -> ta.Mapping[str, ta.Any]:
            return {}

    #

    def __init__(
            self,
            style: Style,
    ) -> None:
        super().__init__()

        self._style = style

    @property
    def style(self) -> Style:
        return self._style

    @dc.dataclass(frozen=True, kw_only=True)
    class CompileResult:
        fn_name: str
        fn_params: ta.Sequence[str]

        hdr_lines: ta.Sequence[str]
        fn_lines: ta.Sequence[str]

        refs: frozenset[Ref]

    @dc.dataclass(frozen=True)
    class _FnParam:
        name: str
        src: str | None = None
        noqa: bool = dc.field(default=False, kw_only=True)

    def _compile_set_attr(
            self,
            attr_name: str,
            value_src: str,
            if_present: IfAttrPresent,
    ) -> list[str]:
        setattr_stmt = f'setattr({CLS_IDENT}, {attr_name!r}, {value_src})'

        if if_present == 'skip':
            return [
                f'if {attr_name!r} not in {CLS_IDENT}.__dict__:',
                f'    {setattr_stmt}',
            ]

        elif if_present == 'replace':
            return [
                setattr_stmt,
            ]

        elif if_present == 'error':
            return [
                f'if {attr_name!r} in {CLS_IDENT}.__dict__:',
                (
                    f'    '
                    f'raise {TYPE_ERROR_GLOBAL.ident}'
                    f'(f"Cannot overwrite attribute {attr_name} in class {{{CLS_IDENT}.__name__}}")'
                ),
                setattr_stmt,
            ]

        else:
            raise ValueError(if_present)

    def compile(
            self,
            fn_name: str,
            ops: ta.Sequence[Op],
    ) -> CompileResult:
        body_lines: list[str] = []

        for i, op in enumerate(ops):
            if i:
                body_lines.append('')

            if isinstance(op, SetAttrOp):
                if isinstance(v := op.value, OpRef):
                    vs = v.ident()
                    body_lines.extend([
                        f'if isinstance({vs}, {FUNCTION_TYPE_GLOBAL.ident}):'
                        f'    {vs}.__qualname__ = f"{{{CLS_IDENT}.__qualname__}}.{{{vs}.__name__}}"',
                    ])
                else:
                    vs = repr(repr_round_trip_value(v))

                body_lines.extend(self._compile_set_attr(
                    op.name,
                    vs,
                    op.if_present,
                ))

            elif isinstance(op, AddMethodOp):
                body_lines.extend([
                    *op.src.splitlines(),
                    f'',
                    f'{op.name}.__qualname__ = f"{{{CLS_IDENT}.__qualname__}}.{op.name}"',
                    *self._compile_set_attr(
                        op.name,
                        op.name,
                        op.if_present,
                    ),
                ])

            elif isinstance(op, AddPropertyOp):
                gen_ident = IDENT_PREFIX + f'property__{op.name}'

                gen_lines = [
                    f'def {gen_ident}():',
                    f'    @{PROPERTY_GLOBAL.ident}',
                    *[
                        f'    {l}'
                        for l in check.not_none(op.get_src).splitlines()
                    ],
                ]
                if op.set_src is not None:
                    gen_lines.extend([
                        f'',
                        f'    @{op.name}.setter',
                        *[
                            f'    {l}'
                            for l in op.set_src.splitlines()
                        ],
                    ])
                if op.del_src is not None:
                    raise NotImplementedError
                gen_lines.extend([
                    f'',
                    f'    return {op.name}',
                ])

                body_lines.extend([
                    *gen_lines,
                    f'',
                    f'setattr({CLS_IDENT}, {op.name!r}, {gen_ident}())',
                ])

            else:
                raise TypeError(op)

        #

        refs = frozenset.union(*[get_op_refs(o) for o in ops])

        params: list[OpCompiler._FnParam] = [
            OpCompiler._FnParam(CLS_IDENT),
            *[
                OpCompiler._FnParam(p)
                for p in sorted(
                    r.ident()
                    for r in refs
                    if isinstance(r, OpRef)
                )
            ],
        ]

        params.extend([
            OpCompiler._FnParam(
                k.ident,
                src=f'{k.ident}={v.src}' if not v.src.startswith('.') else k.ident,
                noqa=k.ident != k.ident.lower() or not v.src.startswith('.'),
            )
            for k, v in sorted(FN_GLOBALS.items(), key=lambda t: t[0])
        ])

        #

        fn_lines = [
            f'def {fn_name}(',
            f'    *,',
            *[
                f'    {p.src if p.src is not None else p.name},{"  # noqa" if p.noqa else ""}'
                for p in params
            ],
            f'):',
            *[
                f'    {l}'
                for l in body_lines
            ],
        ]

        #

        return self.CompileResult(
            fn_name=fn_name,
            fn_params=[p.name for p in params],

            hdr_lines=self._style.header_lines(),
            fn_lines=fn_lines,

            refs=refs,
        )



import dataclasses as dc
import reprlib
import types
import typing as ta

from ...errors import FieldFnValidationError
from ...errors import FieldTypeValidationError
from ...errors import FnValidationError
from .idents import IDENT_PREFIX


##


@ta.final
class FnGlobal(ta.NamedTuple):
    ident: str


class FnGlobalValue(ta.NamedTuple):
    value: ta.Any
    src: str  # Leading '.' denotes a dataclasses-internal symbol


FN_GLOBAL_IMPORTS: ta.Mapping[str, types.ModuleType] = {
    'dataclasses': dc,
    'reprlib': reprlib,
    'types': types,
}


FN_GLOBALS: ta.Mapping[FnGlobal, FnGlobalValue] = {
    (ISINSTANCE_GLOBAL := FnGlobal(IDENT_PREFIX + 'isinstance')): FnGlobalValue(isinstance, 'isinstance'),
    (NONE_GLOBAL := FnGlobal(IDENT_PREFIX + 'None')): FnGlobalValue(None, 'None'),
    (PROPERTY_GLOBAL := FnGlobal(IDENT_PREFIX + 'property')): FnGlobalValue(property, 'property'),
    (TYPE_ERROR_GLOBAL := FnGlobal(IDENT_PREFIX + 'TypeError')): FnGlobalValue(TypeError, 'TypeError'),

    (OBJECT_SETATTR_GLOBAL := FnGlobal(IDENT_PREFIX + 'object_setattr')): FnGlobalValue(
        object.__setattr__,
        'object.__setattr__',
    ),

    (FROZEN_INSTANCE_ERROR_GLOBAL := FnGlobal(IDENT_PREFIX + 'FrozenInstanceError')): FnGlobalValue(
        dc.FrozenInstanceError,
        'dataclasses.FrozenInstanceError',
    ),
    (HAS_DEFAULT_FACTORY_GLOBAL := FnGlobal(IDENT_PREFIX + 'HAS_DEFAULT_FACTORY')): FnGlobalValue(
        dc._HAS_DEFAULT_FACTORY,  # type: ignore[attr-defined]  # noqa
        'dataclasses._HAS_DEFAULT_FACTORY',
    ),
    (MISSING_GLOBAL := FnGlobal(IDENT_PREFIX + 'MISSING')): FnGlobalValue(
        dc.MISSING,  # noqa
        'dataclasses.MISSING',
    ),

    (REPRLIB_RECURSIVE_REPR_GLOBAL := FnGlobal(IDENT_PREFIX + '_recursive_repr')): FnGlobalValue(
        reprlib.recursive_repr,
        'reprlib.recursive_repr',
    ),

    (FUNCTION_TYPE_GLOBAL := FnGlobal(IDENT_PREFIX + 'FunctionType')): FnGlobalValue(
        types.FunctionType,
        'types.FunctionType',
    ),

    (FIELD_FN_VALIDATION_ERROR_GLOBAL := FnGlobal(IDENT_PREFIX + 'FieldFnValidationError')): FnGlobalValue(
        FieldFnValidationError,
        '.errors.FieldFnValidationError',
    ),
    (FIELD_TYPE_VALIDATION_ERROR_GLOBAL := FnGlobal(IDENT_PREFIX + 'FieldTypeValidationError')): FnGlobalValue(
        FieldTypeValidationError,
        '.errors.FieldTypeValidationError',
    ),
    (FN_VALIDATION_ERROR_GLOBAL := FnGlobal(IDENT_PREFIX + 'FnValidationError')): FnGlobalValue(
        FnValidationError,
        '.errors.FnValidationError',
    ),
}


FN_GLOBAL_VALUES: ta.Mapping[str, ta.Any] = {
    k.ident: v.value
    for k, v in FN_GLOBALS.items()
}



import dataclasses as dc
import typing as ta

from ..impl.api.classes.params import get_class_spec
from .specs import get_metaclass_spec


##


CONFER_CLASS_PARAMS: tuple[str, ...] = (
    'eq',
    'frozen',
    'kw_only',

    'reorder',
    'cache_hash',
    'generic_init',
    'override',

    'repr_id',
    'terse_repr',

    'allow_redundant_decorator',
)

CONFER_METACLASS_PARAMS: tuple[str, ...] = (
    'confer',
    'final_subclasses',
    'abstract_immediate_subclasses',
)


def confer_kwarg(out: dict[str, ta.Any], k: str, v: ta.Any) -> None:
    if k in out:
        if out[k] != v:
            raise ValueError
    else:
        out[k] = v


def confer_kwargs(
        bases: ta.Sequence[type],
        kwargs: ta.Mapping[str, ta.Any],
) -> dict[str, ta.Any]:
    out: dict[str, ta.Any] = {}
    for base in bases:
        if not dc.is_dataclass(base):
            continue

        if (bcs := get_class_spec(base)) is None:
            continue

        if not (bmp := get_metaclass_spec(bcs)).confer:
            continue

        for ck in bmp.confer:
            if ck in kwargs:
                continue

            if ck in CONFER_CLASS_PARAMS:
                confer_kwarg(out, ck, getattr(bcs, ck))

            elif ck in CONFER_METACLASS_PARAMS:
                confer_kwarg(out, ck, getattr(bmp, ck))

            else:
                raise KeyError(ck)

    return out



import typing as ta


##


class JmespathError(ValueError):
    pass


class ParseError(JmespathError):
    _ERROR_MESSAGE = 'Invalid jmespath expression'

    def __init__(
            self,
            lex_position,
            token_value,
            token_type,
            msg: str = _ERROR_MESSAGE,
    ) -> None:
        super().__init__(lex_position, token_value, token_type)

        self.lex_position = lex_position
        self.token_value = token_value
        self.token_type = token_type.upper()
        self.msg = msg
        # Whatever catches the ParseError can fill in the full expression
        self.expression: ta.Any = None

    def __str__(self):
        # self.lex_position +1 to account for the starting double quote char.
        underline = ' ' * (self.lex_position + 1) + '^'
        return (
            f'{self.msg}: Parse error at column {self.lex_position}, '
            f'token "{self.token_value}" ({self.token_type}), for expression:\n"{self.expression}"\n{underline}'
        )


class IncompleteExpressionError(ParseError):
    def set_expression(self, expression):
        self.expression = expression
        self.lex_position = len(expression)
        self.token_type = None
        self.token_value = None

    def __str__(self):
        # self.lex_position +1 to account for the starting double quote char.
        underline = ' ' * (self.lex_position + 1) + '^'
        return (
            f'Invalid jmespath expression: Incomplete expression:\n'
            f'"{self.expression}"\n{underline}'
        )


class LexerError(ParseError):
    def __init__(self, lexer_position, lexer_value, message, expression: ta.Any | None = None):
        self.lexer_position = lexer_position
        self.lexer_value = lexer_value
        self.message = message

        super().__init__(lexer_position, lexer_value, message)

        # Whatever catches LexerError can set this.
        self.expression = expression

    def __str__(self):
        underline = ' ' * self.lexer_position + '^'
        return f'Bad jmespath expression: {self.message}:\n{self.expression}\n{underline}'


class ArityError(ParseError):
    def __init__(self, expected, actual, name):
        self.expected_arity = expected
        self.actual_arity = actual
        self.function_name = name
        self.expression = None

    def __str__(self):
        return (
            f"Expected {self.expected_arity} {self._pluralize('argument', self.expected_arity)} "
            f'for function {self.function_name}(), received {self.actual_arity}'
        )

    def _pluralize(self, word, count):
        if count == 1:
            return word
        else:
            return word + 's'


class VariadicArityError(ArityError):
    def __str__(self):
        return (
            f"Expected at least {self.expected_arity} {self._pluralize('argument', self.expected_arity)} "
            f'for function {self.function_name}(), received {self.actual_arity}'
        )


class JmespathTypeError(JmespathError):
    def __init__(
            self,
            function_name,
            current_value,
            actual_type,
            expected_types,
    ):
        self.function_name = function_name
        self.current_value = current_value
        self.actual_type = actual_type
        self.expected_types = expected_types

    def __str__(self):
        return (
            f'In function {self.function_name}(), invalid type for value: {self.current_value}, '
            f'expected one of: {self.expected_types}, received: "{self.actual_type}"'
        )


class JmespathValueError(JmespathError):
    def __init__(
            self,
            function_name,
            current_value,
            expected_types,
    ):
        self.function_name = function_name
        self.current_value = current_value
        self.expected_types = expected_types

    def __str__(self):
        return (
            f'In function {self.function_name}(), invalid value: "{self.current_value}", '
            f'expected: {self.expected_types}'
        )


class EmptyExpressionError(JmespathError):
    def __init__(self):
        super().__init__('Invalid Jmespath expression: cannot be empty.')


class UnknownFunctionError(JmespathError):
    pass


class UndefinedVariableError(JmespathError):
    def __init__(self, varname):
        self.varname = varname

        super().__init__(f'Reference to undefined variable: {self.varname}')



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .elements import Element
from .elements import Elements
from .elements import as_elements


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class Overrides(Element, lang.Final):
    src: Elements = dc.xfield(coerce=check.of_isinstance(Elements))
    ovr: Elements = dc.xfield(coerce=check.of_isinstance(Elements))


def override(src: ta.Any, *ovr: ta.Any) -> Element:
    return Overrides(as_elements(src), as_elements(*ovr))



import typing as ta
import urllib.error
import urllib.request

from .. import check
from .. import dataclasses as dc
from ..formats import json


##


def _tag_sort_key(s: str) -> tuple:
    l = []
    for p in s.split('.'):
        v: ta.Any
        try:
            v = int(p)
        except ValueError:
            v = s
        l.append((not isinstance(v, int), v))
    return tuple(l)


DEFAULT_TAG_SUFFIX_DELIM = '-+'


def split_tag_suffix(
        tag: str,
        suffix_delim: ta.Iterable[str] = DEFAULT_TAG_SUFFIX_DELIM,
) -> tuple[str, str | None]:
    for d in suffix_delim:
        if d in tag:
            p, _, s = tag.partition(d)
            return p, s
    return tag, None


def select_latest_tag(
        tags: ta.Iterable[str],
        *,
        base: str | None = None,
        suffix: str | None = None,
        suffix_delim: ta.Iterable[str] = '-+',
) -> str:
    check.not_isinstance(tags, str)

    tags_by_sfx: dict[str | None, set[tuple[tuple, str]]] = {}
    for t in tags:
        p, s = split_tag_suffix(t, suffix_delim)
        tags_by_sfx.setdefault(s, set()).add((_tag_sort_key(p), t))

    if base is not None:
        bp, bs = split_tag_suffix(base, suffix_delim)
        if suffix is None:
            suffix = bs
        base_key = _tag_sort_key(bp)
    else:
        base_key = None

    sl = sorted(tags_by_sfx[suffix])

    if base_key is not None:
        sl = [(k, t) for k, t in sl if k[0] == base_key[0]]

    return sl[-1][1]


##


@dc.dataclass(frozen=True)
class HubRepoInfo:
    repo: str
    tags: ta.Sequence[str]
    manifests: ta.Mapping[str, ta.Mapping[str, ta.Any]]


def get_hub_repo_info(
        repo: str,
        *,
        auth_url: str = 'https://auth.docker.io/',
        api_url: str = 'https://registry-1.docker.io/v2/',
        tags: ta.Iterable[str] | None = None,
        handled_codes: ta.Container[int] | None = (401, 404),
) -> HubRepoInfo | None:
    """
    https://stackoverflow.com/a/39376254

    ==

    repo=library/nginx
    token=$(
        curl -s "https://auth.docker.io/token?service=registry.docker.io&scope=repository:${repo}:pull" \
        | jq -r '.token' \
    )
    curl -H "Authorization: Bearer $token" -s "https://registry-1.docker.io/v2/${repo}/tags/list" | jq
    curl \
        -H "Accept: application/vnd.docker.distribution.manifest.v2+json" \
        -H "Accept: application/vnd.docker.distribution.manifest.list.v2+json" \
        -H "Authorization: Bearer $token" \
        -s "https://registry-1.docker.io/v2/${repo}/manifests/latest" \
    | jq .
    """
    if tags is not None:
        check.not_isinstance(tags, str)
    else:
        tags = []

    auth_url = auth_url.rstrip('/')
    api_url = api_url.rstrip('/')

    if '/' not in repo:
        repo = '_/' + repo
    if repo.startswith('_/'):
        repo = 'library' + repo[1:]

    #

    def req_json(url: str, **kwargs: ta.Any) -> ta.Any:
        with urllib.request.urlopen(urllib.request.Request(url, **kwargs)) as resp:  # noqa
            return json.loads(resp.read().decode('utf-8'))

    #

    token_dct = req_json(f'{auth_url}/token?service=registry.docker.io&scope=repository:{repo}:pull')
    token = token_dct['token']

    req_hdrs = {'Authorization': f'Bearer {token}'}

    #

    try:
        tags_resp = req_json(
            f'{api_url}/{repo}/tags/list',
            headers=req_hdrs,
        )
    except urllib.error.HTTPError as ue:
        if ue.code in (handled_codes or ()):
            return None
        else:
            raise

    tags_dct = tags_resp.get('tags', {})

    manis = {}
    for tag in tags:
        mani = req_json(
            f'{api_url}/{repo}/manifests/latest',
            headers={
                **req_hdrs,
                'Accept': 'application/vnd.docker.distribution.manifest.v2+json',
            },
        )
        manis[tag] = mani

    return HubRepoInfo(
        repo,
        tags_dct,
        manis,
    )



"""
MIT License

Copyright (c) 2022 Gabriel Ochsenhofer

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import abc
import random
import typing as ta

from ... import lang
from ..sorted.sorted import SortedItems
from . import treap
from .persistent import PersistentMap
from .persistent import PersistentMapping


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class TreapMap(PersistentMap[K, V], SortedItems[K, V]):
    __slots__ = ('_n', '_c')

    def __init__(
            self,
            *,
            _n: treap.TreapNode[tuple[K, V]] | None,
            _c: treap.Comparer[tuple[K, V]],
    ) -> None:
        super().__init__()

        self._n = _n
        self._c = _c

    def __len__(self) -> int:
        # TODO: memo
        # TODO: itertools lol
        if self._n is None:
            return 0
        result = 0
        for _ in self._n:
            result += 1
        return result

    def __contains__(self, item: K) -> bool:
        try:
            self[item]  # noqa
        except KeyError:
            return False
        else:
            return True

    def __getitem__(self, item: K) -> V:
        n = treap.find(self._n, (item, None), self._c)  # type: ignore
        if n is None:
            raise KeyError(item)
        return n.value

    def __iter__(self) -> ta.Iterator[K]:
        i = self.items()
        while i.has_next():
            yield i.next()[0]

    def items(self) -> 'TreapMapIterator[K, V]':
        i = TreapMapIterator(
            _st=[],
            _n=self._n,
        )
        while (n := i._n) is not None and n.left is not None:  # noqa
            i._st.append(n)  # noqa
            i._n = n.left  # noqa
        return i

    def items_desc(self) -> 'TreapMapReverseIterator[K, V]':
        i = TreapMapReverseIterator(
            _st=[],
            _n=self._n,
        )
        while (n := i._n) is not None and n.right is not None:  # noqa
            i._st.append(n)  # noqa
            i._n = n.right  # noqa
        return i

    def items_from(self, k: K) -> 'TreapMapIterator[K, V]':
        lst = treap.place(self._n, (k, None), self._c)  # type: ignore
        i = TreapMapIterator(
            _st=lst,
            _n=lst.pop(),
        )
        return i

    def items_from_desc(self, k: K) -> 'TreapMapReverseIterator[K, V]':
        lst = treap.place(self._n, (k, None), self._c, desc=True)  # type: ignore
        i = TreapMapReverseIterator(
            _st=lst,
            _n=lst.pop(),
        )
        return i

    def with_(self, k: K, v: V) -> 'TreapMap[K, V]':
        node = treap.TreapNode(
            _value=(k, v),
            _priority=int(random.random() * 0xFFFFFFFF),
            _left=None,
            _right=None,
        )
        n = treap.union(self._n, node, self._c, True)
        return TreapMap(_n=n, _c=self._c)

    def without(self, k: K) -> 'TreapMap[K, V]':
        n = treap.delete(self._n, (k, None), self._c)  # type: ignore
        return TreapMap(_n=n, _c=self._c)

    def default(self, k: K, v: V) -> 'TreapMap[K, V]':
        try:
            self[k]  # noqa
        except KeyError:
            return self.with_(k, v)
        else:
            return self


def new_treap_map(cmp: ta.Callable[[tuple[K, V], tuple[K, V]], int]) -> PersistentMap[K, V]:
    return TreapMap(_n=None, _c=cmp)


#


class TreapDict(TreapMap[K, V], PersistentMapping[K, V]):
    __contains__ = TreapMap.__contains__
    items = TreapMap.items


def new_treap_dict(cmp: ta.Callable[[tuple[K, V], tuple[K, V]], int]) -> PersistentMapping[K, V]:
    return TreapDict(_n=None, _c=cmp)


##


class BaseTreapMapIterator(lang.Abstract, ta.Iterator[tuple[K, V]], ta.Generic[K, V]):
    __slots__ = ('_st', '_n')

    def __init__(
            self,
            *,
            _st: list[treap.TreapNode[tuple[K, V]]],
            _n: treap.TreapNode[tuple[K, V]] | None,
    ) -> None:
        super().__init__()

        self._st = _st
        self._n = _n

    def __iter__(self) -> ta.Self:
        return self

    def __next__(self) -> tuple[K, V]:
        return self.next()

    def has_next(self) -> bool:
        return self._n is not None

    @abc.abstractmethod
    def next(self) -> tuple[K, V]:
        raise NotImplementedError


class TreapMapIterator(BaseTreapMapIterator[K, V]):
    __slots__ = BaseTreapMapIterator.__slots__

    def next(self) -> tuple[K, V]:
        n = self._n
        if n is None:
            raise StopIteration
        if n.right is not None:
            self._n = n.right
            while self._n.left is not None:
                self._st.append(self._n)
                self._n = self._n.left
        elif len(self._st) > 0:
            self._n = self._st[-1]
            self._st.pop()
        else:
            self._n = None
        return n.value


class TreapMapReverseIterator(BaseTreapMapIterator[K, V]):
    __slots__ = BaseTreapMapIterator.__slots__

    def next(self) -> tuple[K, V]:
        n = self._n
        if n is None:
            raise StopIteration
        if n.left is not None:
            self._n = n.left
            while self._n.right is not None:
                self._st.append(self._n)
                self._n = self._n.right
        elif len(self._st) > 0:
            self._n = self._st[-1]
            self._st.pop()
        else:
            self._n = None
        return n.value



# ruff: noqa: I001
from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from . import api  # noqa

    from . import dbapi  # noqa

    from . import queries  # noqa

    #

    from .dbs import (  # noqa
        DbType,
        DbTypes,

        DbLoc,
        UrlDbLoc,
        HostDbLoc,

        DbSpec,
    )

    from .params import (  # noqa
        ParamKey,
        ParamsPreparer,
        LinearParamsPreparer,
        NumericParamsPreparer,
        NamedParamsPreparer,

        ParamStyle,
        make_params_preparer,

        substitute_params,
    )

    from .qualifiedname import (  # noqa
        QualifiedName,
        qn,
    )



# @omlish-lite
# ruff: noqa: UP045
import errno
import socket
import typing as ta
import urllib.parse

from ....io.buffers import ReadableListBuffer
from ....lite.check import check
from ...coro.client.connection import CoroHttpClientConnection
from ...coro.client.response import CoroHttpClientResponse
from ...coro.io import CoroHttpIo
from ...headers import HttpHeaders
from ...urls import unparse_url_request_path
from ..base import HttpClientContext
from ..base import HttpClientError
from ..base import HttpRequest
from ..sync import HttpClient
from ..sync import StreamHttpResponse


T = ta.TypeVar('T')


##


class CoroHttpClient(HttpClient):
    class _Connection:
        def __init__(self, req: HttpRequest) -> None:
            super().__init__()

            self._req = req
            self._ups = urllib.parse.urlparse(req.url)

            self._ssl = self._ups.scheme == 'https'

        _cc: ta.Optional[CoroHttpClientConnection] = None
        _resp: ta.Optional[CoroHttpClientResponse] = None

        _sock: ta.Optional[socket.socket] = None
        _sock_file: ta.Optional[ta.BinaryIO] = None

        _ssl_context: ta.Any = None

        #

        def _create_https_context(self, http_version: int) -> ta.Any:
            # https://github.com/python/cpython/blob/a7160912274003672dc116d918260c0a81551c21/Lib/http/client.py#L809
            import ssl

            # Function also used by urllib.request to be able to set the check_hostname attribute on a context object.
            context = ssl.create_default_context()

            # Send ALPN extension to indicate HTTP/1.1 protocol.
            if http_version == 11:
                context.set_alpn_protocols(['http/1.1'])

            # Enable PHA for TLS 1.3 connections if available.
            if context.post_handshake_auth is not None:
                context.post_handshake_auth = True

            return context

        #

        def setup(self) -> StreamHttpResponse:
            check.none(self._sock)
            check.none(self._ssl_context)

            self._cc = cc = CoroHttpClientConnection(
                check.not_none(self._ups.hostname),
                default_port=CoroHttpClientConnection.HTTPS_PORT if self._ssl else CoroHttpClientConnection.HTTP_PORT,
            )

            if self._ssl:
                self._ssl_context = self._create_https_context(self._cc.http_version)

            try:
                self._run_coro(cc.connect())

                self._run_coro(cc.request(
                    self._req.method or 'GET',
                    unparse_url_request_path(self._ups) or '/',
                    self._req.data,
                    hh.single_str_dct if (hh := self._req.headers_) is not None else {},
                ))

                self._resp = resp = self._run_coro(cc.get_response())

                return StreamHttpResponse(
                    status=resp._state.status,  # noqa
                    headers=HttpHeaders(resp._state.headers.items()),  # noqa
                    request=self._req,
                    underlying=self,
                    _stream=ReadableListBuffer().new_buffered_reader(self),
                    _closer=self.close,
                )

            except Exception:
                self.close()
                raise

        def _run_coro(self, g: ta.Generator[ta.Any, ta.Any, T]) -> T:
            i = None

            while True:
                try:
                    o = g.send(i)
                except StopIteration as e:
                    return e.value

                try:
                    i = self._handle_io(o)
                except OSError as e:
                    raise HttpClientError from e

        def _handle_io(self, o: CoroHttpIo.Io) -> ta.Any:
            if isinstance(o, CoroHttpIo.ConnectIo):
                check.none(self._sock)
                self._sock = socket.create_connection(*o.args, **(o.kwargs or {}))

                if self._ssl_context is not None:
                    self._sock = self._ssl_context.wrap_socket(
                        self._sock,
                        server_hostname=check.not_none(o.server_hostname),
                    )

                # Might fail in OSs that don't implement TCP_NODELAY
                try:
                    self._sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                except OSError as e:
                    if e.errno != errno.ENOPROTOOPT:
                        raise

                self._sock_file = self._sock.makefile('rb')

                return None

            elif isinstance(o, CoroHttpIo.CloseIo):
                check.not_none(self._sock).close()
                return None

            elif isinstance(o, CoroHttpIo.WriteIo):
                check.not_none(self._sock).sendall(o.data)
                return None

            elif isinstance(o, CoroHttpIo.ReadIo):
                if (sz := o.sz) is not None:
                    return check.not_none(self._sock_file).read(sz)
                else:
                    return check.not_none(self._sock_file).read()

            elif isinstance(o, CoroHttpIo.ReadLineIo):
                return check.not_none(self._sock_file).readline(o.sz)

            else:
                raise TypeError(o)

        def read1(self, n: int = -1, /) -> bytes:
            return self._run_coro(check.not_none(self._resp).read(n if n >= 0 else None))

        def close(self) -> None:
            if self._resp is not None:
                self._resp.close()
            if self._sock is not None:
                self._sock.close()

    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> StreamHttpResponse:
        conn = CoroHttpClient._Connection(req)
        return conn.setup()



import dataclasses as dc
import typing as ta

from ... import codecs
from ..coro import buffer_bytes_stepped_reader_coro
from .base import Compression
from .base import IncrementalCompression


##


@dc.dataclass(frozen=True)
class CompressionEagerCodec(codecs.EagerCodec[bytes, bytes]):
    compression: Compression

    def encode(self, i: bytes) -> bytes:
        return self.compression.compress(i)

    def decode(self, o: bytes) -> bytes:
        return self.compression.decompress(o)


##


@dc.dataclass(frozen=True)
class CompressionIncrementalCodec(codecs.IncrementalCodec[bytes, bytes]):
    compression: IncrementalCompression

    def encode_incremental(self) -> ta.Generator[bytes | None, bytes]:
        return self.compression.compress_incremental()

    def decode_incremental(self) -> ta.Generator[bytes | None, bytes]:
        return buffer_bytes_stepped_reader_coro(self.compression.decompress_incremental())


##


class CompressionCodec(codecs.Codec):
    pass


def make_compression_codec(
        name: str,
        cls: type[Compression],
        *,
        aliases: ta.Sequence[str] | None = None,
) -> CompressionCodec:
    return CompressionCodec(
        name=name,
        aliases=aliases,

        input=bytes,
        output=bytes,

        new=lambda *args, **kwargs: CompressionEagerCodec(cls(*args, **kwargs)),

        new_incremental=(
            lambda *args, **kwargs: CompressionIncrementalCodec(cls(*args, **kwargs))  # noqa
        ) if issubclass(cls, IncrementalCompression) else None,
    )


##


def make_compression_lazy_loaded_codec(
        module: str,
        attr: str,
        codec: CompressionCodec,
) -> codecs.LazyLoadedCodec:
    return codecs.LazyLoadedCodec.new(
        module,
        attr,
        codec,
    )



# ruff: noqa: UP006 UP045
# @omlish-lite
import logging
import typing as ta


LogLevel = int  # ta.TypeAlias


##


@ta.final
class NamedLogLevel(int):
    # logging.getLevelNamesMapping (or, as that is unavailable <3.11, logging._nameToLevel) includes the deprecated
    # aliases.
    _NAMES_BY_INT: ta.ClassVar[ta.Mapping[LogLevel, str]] = dict(sorted(logging._levelToName.items(), key=lambda t: -t[0]))  # noqa

    _INTS_BY_NAME: ta.ClassVar[ta.Mapping[str, LogLevel]] = {v: k for k, v in _NAMES_BY_INT.items()}

    _NAME_INT_PAIRS: ta.ClassVar[ta.Sequence[ta.Tuple[str, LogLevel]]] = list(_INTS_BY_NAME.items())

    #

    _CACHE: ta.ClassVar[ta.MutableMapping[int, 'NamedLogLevel']] = {}

    @ta.overload
    def __new__(cls, name: str, offset: int = 0, /) -> 'NamedLogLevel':
        ...

    @ta.overload
    def __new__(cls, i: int, /) -> 'NamedLogLevel':
        ...

    def __new__(cls, x, offset=0, /):
        if isinstance(x, str):
            return cls(cls._INTS_BY_NAME[x.upper()] + offset)
        elif not offset and (c := cls._CACHE.get(x)) is not None:
            return c
        else:
            return super().__new__(cls, x + offset)

    #

    _name_and_offset: ta.Tuple[str, int]

    @property
    def name_and_offset(self) -> ta.Tuple[str, int]:
        try:
            return self._name_and_offset
        except AttributeError:
            pass

        if (n := self._NAMES_BY_INT.get(self)) is not None:
            t = (n, 0)
        else:
            for n, i in self._NAME_INT_PAIRS:  # noqa
                if self >= i:
                    t = (n, (self - i))
                    break
            else:
                t = ('NOTSET', int(self))

        self._name_and_offset = t
        return t

    @property
    def exact_name(self) -> ta.Optional[str]:
        n, o = self.name_and_offset
        return n if not o else None

    @property
    def effective_name(self) -> str:
        n, _ = self.name_and_offset
        return n

    #

    def __str__(self) -> str:
        return self.exact_name or f'{self.effective_name}{int(self):+}'

    def __repr__(self) -> str:
        n, o = self.name_and_offset
        return f'{self.__class__.__name__}({n!r}{f", {int(o)}" if o else ""})'

    #

    CRITICAL: ta.ClassVar['NamedLogLevel']
    ERROR: ta.ClassVar['NamedLogLevel']
    WARNING: ta.ClassVar['NamedLogLevel']
    INFO: ta.ClassVar['NamedLogLevel']
    DEBUG: ta.ClassVar['NamedLogLevel']
    NOTSET: ta.ClassVar['NamedLogLevel']


NamedLogLevel.CRITICAL = NamedLogLevel(logging.CRITICAL)
NamedLogLevel.ERROR = NamedLogLevel(logging.ERROR)
NamedLogLevel.WARNING = NamedLogLevel(logging.WARNING)
NamedLogLevel.INFO = NamedLogLevel(logging.INFO)
NamedLogLevel.DEBUG = NamedLogLevel(logging.DEBUG)
NamedLogLevel.NOTSET = NamedLogLevel(logging.NOTSET)


NamedLogLevel._CACHE.update({i: NamedLogLevel(i) for i in NamedLogLevel._NAMES_BY_INT})  # noqa



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .dispatch import (  # noqa
        Dispatcher,
    )

    from .functions import (  # noqa
        function,
    )

    from .methods import (  # noqa
        install_method,
        method,
    )



import typing as ta

from ... import cached
from ... import dataclasses as dc
from ..binder import bind
from ..elements import Elemental
from ..keys import Key
from ..keys import as_key
from ..privates import private


##


@ta.final
class WrapperBinderHelper:
    def __init__(self, key: ta.Any) -> None:
        self._key = as_key(key)
        self._root = WrapperBinderHelper._Root()
        self._top = WrapperBinderHelper._Level(self._root, 0)

    @dc.dataclass(frozen=True, eq=False)
    @dc.extra_class_params(repr_id=True)
    class _Root:
        pass

    @dc.dataclass(frozen=True)
    class _Level:
        root: 'WrapperBinderHelper._Root'
        level: int

        def next(self) -> 'WrapperBinderHelper._Level':
            return WrapperBinderHelper._Level(self.root, self.level + 1)

        @cached.property
        def key(self) -> Key:
            return Key(ta.Any, tag=self)

    @property
    def top(self) -> Key:
        return self._top.key

    def push_bind(self, **kwargs: ta.Any) -> Elemental:
        prv = self._top
        nxt = prv.next()
        out = private(
            *([bind(self._key, to_key=prv.key)] if prv.level else []),
            bind(nxt.key, **kwargs, expose=True),
        )
        self._top = nxt
        return out


wrapper_binder_helper = WrapperBinderHelper



import dataclasses as dc
import typing as ta

from .... import check
from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.ops import SetAttrOp
from ..generation.registry import register_generator_type
from ..generation.utils import build_attr_tuple_body_src_lines
from ..processing.base import ProcessingContext
from .fields import InstanceFields


##


HashAction: ta.TypeAlias = ta.Literal['set_none', 'add', 'exception']


# See https://bugs.python.org/issue32929#msg312829 for an if-statement version of this table.
HASH_ACTIONS: ta.Mapping[tuple[bool, bool, bool, bool], HashAction | None] = {
    #
    # +-------------------------------------- unsafe_hash?
    # |      +------------------------------- eq?
    # |      |      +------------------------ frozen?
    # |      |      |      +----------------  has-explicit-hash?
    # v      v      v      v
    (False, False, False, False): None,
    (False, False, False, True): None,
    (False, False, True, False): None,
    (False, False, True, True): None,
    (False, True, False, False): 'set_none',
    (False, True, False, True): None,
    (False, True, True, False): 'add',
    (False, True, True, True): None,
    (True, False, False, False): 'add',
    (True, False, False, True): 'exception',
    (True, False, True, False): 'add',
    (True, False, True, True): 'exception',
    (True, True, False, False): 'add',
    (True, True, False, True): 'exception',
    (True, True, True, False): 'add',
    (True, True, True, True): 'exception',
}


def _raise_hash_action_exception(cls: type) -> ta.NoReturn:
    raise TypeError(f'Cannot overwrite attribute __hash__ in class {cls.__name__}')


CACHED_HASH_ATTR = '__dataclass_hash__'


#


@dc.dataclass(frozen=True)
class HashPlan(Plan):
    action: HashAction

    _: dc.KW_ONLY

    fields: tuple[str, ...] | None = None
    cache: bool | None = None


@register_generator_type(HashPlan)
class HashGenerator(Generator[HashPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[HashPlan] | None:
        class_hash = ctx.cls.__dict__.get('__hash__', dc.MISSING)
        has_explicit_hash = not (class_hash is dc.MISSING or (class_hash is None and '__eq__' in ctx.cls.__dict__))

        action = HASH_ACTIONS[(
            bool(ctx.cs.unsafe_hash),
            bool(ctx.cs.eq),
            bool(ctx.cs.frozen),
            has_explicit_hash,
        )]

        if action == 'set_none':
            return PlanResult(HashPlan(action))  # noqa

        elif action == 'exception':
            _raise_hash_action_exception(ctx.cls)

        elif action == 'add':
            fields = tuple(
                f.name
                for f in ctx[InstanceFields]
                if (f.compare if f.hash is None else f.hash)
            )

            return PlanResult(HashPlan(
                'add',
                fields=fields,
                cache=ctx.cs.cache_hash,
            ))

        elif action is None:
            return None

        else:
            raise ValueError(action)

    def generate(self, pl: HashPlan) -> ta.Iterable[Op]:
        if pl.action == 'set_none':
            return [SetAttrOp('__hash__', None, if_present='replace')]

        elif pl.action != 'add':
            raise ValueError(pl.action)

        lines = [
            'def __hash__(self):',
        ]

        hash_lines: list[str]
        if pl.fields:
            hash_lines = [
                'hash((',
                *build_attr_tuple_body_src_lines(
                    'self',
                    *check.not_none(pl.fields),
                    prefix='    ',
                ),
                '))',
            ]
        else:
            hash_lines = ['hash(())']

        if pl.cache:
            lines.extend([
                f'    try:',
                f'        return self.{CACHED_HASH_ATTR}',
                f'    except AttributeError:',
                f'        pass',
                f'    object.__setattr__(',
                f'        self,',
                f'        {CACHED_HASH_ATTR!r},',
                f'        h := {hash_lines[0]}',
                *[
                    f'        {l}'
                    for l in hash_lines[1:]
                ],
                f'    )',
                f'    return h',
            ])
        else:
            lines.extend([
                f'    return {hash_lines[0]}',
                *[
                    f'    {l}'
                    for l in hash_lines[1:]
                ],
            ])

        return [
            AddMethodOp(
                '__hash__',
                '\n'.join(lines),
                if_present='replace',
            ),
        ]
